#include "Graphics.h"

#include "Logger.h"

#include <string>

using namespace Microsoft::WRL;

#pragma comment (lib, "d3d12.lib")
#pragma comment (lib, "d3d11.lib")
#pragma comment (lib, "d2d1.lib")
#pragma comment (lib, "dwrite.lib")
#pragma comment (lib, "d3dcompiler.lib")
#pragma comment (lib, "dxgi.lib")

D12Graphics::D12Graphics() {
}

D12Graphics::~D12Graphics() {
}

bool D12Graphics::GetResolutions(Config &cfg) {
	IDXGIOutput* adapterOutput{};

	HRESULT result{};
	int count{};

	while (result = InterfaceList[GPU_Num].Adapter->EnumOutputs(count, &adapterOutput) != DXGI_ERROR_NOT_FOUND) {
		if (FAILED(result)) {
			Log_Write("enum output failed\n");

			return false;
		}

		Screen scr{};

		ErrorHR(adapterOutput->GetDesc(&scr.desc));

		DEVMODE displayMode{};
		displayMode.dmSize = sizeof(DEVMODE);

		Error(EnumDisplaySettings(scr.desc.DeviceName, ENUM_CURRENT_SETTINGS, &displayMode));

		float proc = (float)displayMode.dmPelsHeight / (scr.desc.DesktopCoordinates.bottom - scr.desc.DesktopCoordinates.top);

		unsigned int numModes{};
		ErrorHR(adapterOutput->GetDisplayModeList(DXGI_FORMAT_B8G8R8A8_UNORM, DXGI_ENUM_MODES_INTERLACED, &numModes, NULL));

		DXGI_MODE_DESC* displayModeList = new DXGI_MODE_DESC[numModes];
		Error(displayModeList);
		
		auto hr = ErrorHR(adapterOutput->GetDisplayModeList(DXGI_FORMAT_B8G8R8A8_UNORM, DXGI_ENUM_MODES_INTERLACED, &numModes, displayModeList));
		if (FAILED(hr)) { delete[] displayModeList;  return false; }
		
		scr.OutputNum = count;
		scr.DPI = proc * 100;
		cfg.Feature = InterfaceList[GPU_Num].D3D_Level;

		if (cfg.Width != displayModeList[cfg.GraphModeNum].Width || cfg.Height != displayModeList[cfg.GraphModeNum].Height) {
			scr.ModeNum = numModes - 1;
			scr.Width = displayModeList[scr.ModeNum].Width;
			scr.Height = displayModeList[scr.ModeNum].Height;
			cfg.Width = scr.Width;
			cfg.Height = scr.Height;
			cfg.GraphModeNum = scr.ModeNum;
			Log_Write("Default configuration");
		}
		else {
			scr.Width = cfg.Width;
			scr.Height = cfg.Height;
			Log_Write("Configuration file loaded");
		}

		CurrentOutput.push_back(scr);

		delete[] displayModeList;

		count++;
	}

	return true;
}

void D12Graphics::AdapterList(D12Interface& Current) {
	size_t i{};
	for (auto &e : InterfaceList) {
		std::wcout << i <<L": " << e.GPU_Name.c_str() << std::endl;
		i++;
	}
}

bool D12Graphics::CheckVersionSupport() {
	ID3D12Device* pDevice{};
	ErrorHR(D3D12CreateDevice(InterfaceList[GPU_Num].Adapter.Get(), D3D_FEATURE_LEVEL_11_0, _uuidof(ID3D12Device), (void**)&pDevice));
	
	static const D3D_FEATURE_LEVEL fl[] =
	{
		D3D_FEATURE_LEVEL_12_1,
		D3D_FEATURE_LEVEL_12_0,
		D3D_FEATURE_LEVEL_11_1,
		D3D_FEATURE_LEVEL_11_0,
	};

	D3D12_FEATURE_DATA_FEATURE_LEVELS featLevels =
	{
		_countof(fl), fl, D3D_FEATURE_LEVEL_11_0
	};

	ErrorHR(pDevice->CheckFeatureSupport(D3D12_FEATURE_FEATURE_LEVELS, &featLevels, sizeof(featLevels)));

	InterfaceList[GPU_Num].D3D_Level = featLevels.MaxSupportedFeatureLevel;

	switch (featLevels.MaxSupportedFeatureLevel) {
	case D3D_FEATURE_LEVEL_12_1:
		InterfaceList[GPU_Num].strLevel = L"Dx12_1 version\n";
		break;
	case D3D_FEATURE_LEVEL_12_0:
		InterfaceList[GPU_Num].strLevel = L"Dx12_0 version\n";
		break;
	case D3D_FEATURE_LEVEL_11_1:
		InterfaceList[GPU_Num].strLevel = L"Dx11_1 version\n";
		break;
	case D3D_FEATURE_LEVEL_11_0:
		InterfaceList[GPU_Num].strLevel = L"Dx11_0 version\n";
		break;
	}
	
	pDevice->Release();

	return true;
}

bool D12Graphics::CheckSupport() {
	IDXGIFactory4* factory{ nullptr };

	ErrorHR(CreateDXGIFactory1(IID_PPV_ARGS(&factory)));
		
	DXGI_ADAPTER_DESC1 desc;

	int cDevice{ 0 };
	D12Interface Adapt{};

	while (factory->EnumAdapters1(cDevice, &Adapt.Adapter) != DXGI_ERROR_NOT_FOUND) {
		Adapt.Adapter->GetDesc1(&desc);

		if (desc.Flags & DXGI_ADAPTER_FLAG_SOFTWARE)
		{
			factory->EnumWarpAdapter(IID_PPV_ARGS(&Adapt.Adapter));

			ErrorHR(D3D12CreateDevice(Adapt.Adapter.Get(), D3D_FEATURE_LEVEL_11_0, _uuidof(ID3D12Device), nullptr));
			
			Adapt.Hardware = false;
			Adapt.GPU_Name = desc.Description;

			InterfaceList.push_back(Adapt);
		}
		else {
			ErrorHR(D3D12CreateDevice(Adapt.Adapter.Get(), D3D_FEATURE_LEVEL_11_0, _uuidof(ID3D12Device), nullptr));

			Adapt.Hardware = true;
			Adapt.GPU_Name = desc.Description;

			InterfaceList.push_back(Adapt);
		}

		ZeroMemory(&Adapt, sizeof(D12Interface));

		cDevice++;
	}

	if (cDevice == 0) return false;

	return true;
}